/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 */

package org.enhydra.barracuda.discRack.data.disc;

import com.lutris.dods.builder.generator.query.*;
import com.lutris.dods.builder.generator.dataobject.GenericDO;

import com.lutris.appserver.server.*;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.session.*;

import java.io.*;
import java.sql.*;
import java.util.*;
import java.math.*;

import java.util.Date;  // when I say Date, I don't mean java.sql.Date

/**
 * DiscQuery is used to query the Disc table in the database.<BR>
 * It returns objects of type DiscDO.
 * <P>
 * General usage:
 * <P>
 *     In DODS:
 *        Create a Data Object named "Dog",
 *        and create an Attribute named "Name",
 *        and set that Attribute to "Can be queried."
 *        DODS will then generate the method DogQuery.setQueryName().
 * <P>
 *     In your Business Layer, prepare the query:<BR>
 * <P><PRE>
 *             DogQuery dq = new DogQuery();
 *             dq.setQueryName("Rex")
 *             if ( Rex is a reserved name )
 *                 dq.requireUniqueInstance();
 * </PRE>
 *     Then, get the query results one of two ways:
 * <P>
 *         #1:<PRE>
 *             String names = "";
 *             DogDO[] dogs = dq.getDOArray();
 *             for ( int i = 0; i < dogs.length; i++ ) {
 *                 names += dogs[i].getName() + " ";
 *             }
 * </PRE>
 *      or #2:<PRE>
 *             String names = "";
 *             DogDO dog;
 *             while ( null != ( dog = dq.getNextDO() ) ) {
 *                 names += dog.getName() + " ";
 *             }
 * </PRE>
 * <P>
 *     Note:   If <CODE>requireUniqueInstance()</CODE> was called,
 *             then <CODE>getDOArray()</CODE> or <CODE>getNextDO()</CODE>
 *             will throw an exception if more than one "Rex" was found.
 * <P>
 *     Note:   Results of the query will come from the Data Object cache if:
 *             -  The cache is available.
 *             -  Matches were found in the cache.
 *             -  No other tables (Data Objects of other types) were involved
 *                in the query.
 *                This can happen if you extend the <CODE>DogQuery</CODE> class
 *                and you make calls to the <CODE>QueryBuilder</CODE> object
 *                to add SQL involving other tables.
 *             If any of these conditions is not true,
 *             then any results from the query will come from the database.
 * <P>
 *     To reuse the query object, call:
 * <P><PRE>
 *             dq.reset();
 * </PRE>
 * @author ChristianC
 * @version $Revision: 1.2 $
 */

public class DiscQuery


	implements Query


{
    private Connection conn = null;

    /**
     * Public constructor.
     */

    public DiscQuery() {
	builder = new QueryBuilder( "Disc", "Disc.*" );
	builder.setDatabaseVendor( "Standard" );
	builder.setStringMatchDetails( "LIKE", "%" );
	reset();
    }


    private ResultSet		resultSet	= null;
    private DiscDO[]	DOs		= null;
    private int			arrayIndex	= -1;
    private boolean		needToRun	= true;
    private boolean		doneResultSet	= true;
    private boolean 		uniqueInstance	= false;
    private Vector		cacheHits	= null;
    private DBQuery		dbQuery		= null;
    private static final boolean partialCache = false;


    private QueryBuilder	builder		= null;


    private Connection getConnection()
    throws DataObjectException {
	String task = "";
        try {

	    task = "DiscDO.createConnection()";
            Connection conn = DiscDO.createConnection();

	    if ( null == conn )
	    	throw new DataObjectException( task + " returned null." );
	    return conn;
	} catch ( Exception e ) {
	    reset();
	    throw new DataObjectException(
		"ERROR: DiscQuery.getConnection() : ", e );
        }
    }

    /**
     *
     *
     * @return the
     * @param hitDb if true, query will ignore cache (if any)
     */
    private boolean hitDb = true;
    public void hitDatabase() { hitDb = true; }

    /**
     * Perform the query on the database, and prepare the
     * array of returned DO objects.
     *
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */
    private void runQuery()
    throws DataObjectException, NonUniqueQueryException {
	needToRun = false;
	arrayIndex = -1;
	doneResultSet = false;

	try {
	    Vector results;
	    if ( partialCache && 0 == cacheHits.size() )
		hitDb = true;
	    if ( hitDb ) {


		//dbQuery = Enhydra.getDatabaseManager().createQuery();
		dbQuery = DiscDO.createQuery();
		dbQuery.query( this );    // invokes executeQuery



		results = new Vector();
		if ( ! oneAtATime ) {
		    while ( resultSet.next() ) {
			results.addElement(
			    DiscDO.createExisting ( resultSet ) );
		    }
		    doneResultSet = true;
		}
	    } else {
		results = cacheHits;	 // executeQuery saw cache hits
	    }

	    if ( results.size() > 1 && uniqueInstance )
		throw new NonUniqueQueryException(
		    "Too many rows returned from database" );
	    DOs = new DiscDO [ results.size() ];
	    for ( int i = 0; i < results.size(); i++ ) {
		DOs[i] = ( DiscDO ) results.elementAt( i );
	    }
	    arrayIndex = 0;
	} catch ( SQLException se ) {
	    if (null == se.getSQLState() ) {
		throw new DataObjectException(
		    "Unknown SQLException", se );
	    }
	    if (	se.getSQLState().startsWith("02") &&
			se.getErrorCode() == 100 ) {
		throw new DataObjectException(
		    "Update or delete DO is out of synch", se );
	    } else if (	se.getSQLState().equals("S1000") &&
			se.getErrorCode() == -268 ) {
		throw new DataObjectException(
		    "Integrity constraint violation", se );
	    } else {
		throw new DataObjectException( "Data Object Error", se );
	    }
	} catch ( ObjectIdException oe ) {
	    throw new DataObjectException( "Object ID Error", oe );
	} catch ( DatabaseManagerException de ) {
	    throw new DataObjectException( "Database connection Error", de );

	} finally {
	    release();

	}
    }

    /**
     * Limit the number of rows (DOs) returned.
     * NOTE: When setting a limit on rows returned by a query,
     * you usually want to use a call to an addOrderBy method
     * to cause the most interesting rows to be returned first.
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */
    public void setMaxRows( int maxRows )
    throws DataObjectException, NonUniqueQueryException {

	hitDb = true;
	builder.setMaxRows( maxRows );

    }


    /**
     * Force getNextDO to read one row at a time from the ResultSet.
     * Normally, getNextDO/getDOArray will read all rows into a private array.
     * When huge numbers of rows are expected, calling disableBuffering
     * will avoid OutOfMemory conditions.
     *
     * NOTE: if you call disableBuffering(), and you do not retrieve
     * all the DOs (that is, your last call to getNextDO does not return null)
     * then you should call reset() to free the Connection.
     *
     * NOTE: if you call disableBuffering(), do not call getDOArray;
     * it will throw an Exception.
     *
     * @return the next DiscDO object found by this DiscQuery.
     */
    public void disableBuffering() {
	oneAtATime = true;
    }
    private boolean oneAtATime = false;


    /**
     * Return array of DOs constructed from ResultSet returned by query.
     * @param conn the Connection object to use.
     * The caller is responsible for closing the connection.
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */

    /**
     * Return array of DOs constructed from ResultSet returned by query.
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */
    public DiscDO[] getDOArray()
    throws DataObjectException, NonUniqueQueryException {
	if ( oneAtATime )
	    throw new DataObjectException(
		"Cannot call getDOArray() after disableBuffering()." );
	if ( needToRun )
	    runQuery();
	return DOs;
    }


    /**
     * Return successive DOs from array built from ResultSet returned by query.
     * @param conn the Connection object to use.
     * The caller is responsible for closing the connection.
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */

    /**
     * Return successive DOs from array built from ResultSet returned by query.
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */
    public DiscDO getNextDO()
    throws DataObjectException, NonUniqueQueryException {
	return _getNextDO();
    }


    /**
     * Called by public getNextDO methods.
     * @exception DataObjectException If a database access error occurs.
     * @exception NonUniqueQueryException If too many rows were found.
     */
    private DiscDO _getNextDO()
    throws DataObjectException, NonUniqueQueryException {
	if ( needToRun )
	    runQuery();
	if ( oneAtATime ) {
            try {
		if ( resultSet.next() ) {
		    return DiscDO.createExisting ( resultSet );
		} else {
		    reset();
		    return null;
		}
            } catch ( SQLException e ) {
		reset();
                throw new DataObjectException(
		    "Could not create DiscDO: ", e );
	    } catch ( ObjectIdException oe ) {
		reset();
		throw new DataObjectException( "Object ID Error", oe );
	    } catch ( DatabaseManagerException de ) {
		reset();
		throw new DataObjectException( "Database connection Error", de);
	    }
	} else {
	    if ( null == DOs ) {
		/* This should never happen.
		 * runQuery() should either throw an exception
		 * or create an array of DOs, possibly of zero length.
		 */
		return null;
	    }
	    if ( arrayIndex < DOs.length )
		return DOs[ arrayIndex++ ];
	    return null;
	}
    }



    public int deleteFoundRows()
    throws DataObjectException {
	int ret = 0;
        try {
            ret = builder.executeDelete( getConnection() );
	} catch ( Exception e ) {
	    reset();
	    throw new DataObjectException(
		"Error deleting rows from Disc: ", e );
        }
	return ret;
    }

    public int updateFoundRows()
    throws DataObjectException {
	int ret = 0;
        try {
            ret = builder.executeUpdate( getConnection() );
	} catch ( Exception e ) {
	    reset();
	    throw new DataObjectException(
		"Error updating rows in Disc: ", e );
        }
	return ret;
    }



    /**
     * Set the OID to query.
     * WARNING!  This method assumes that table <CODE>Disc</CODE>
     * has a column named <CODE>"oid"</CODE>.
     * This method is called from the DO classes to retrieve an object by id.
     *
     * @param oid The object id to query.
     */
    public void setQueryOId(ObjectId oid) {
        // Remove from cacheHits any DOs that do not meet this
        // setQuery requirement.
	if ( null == oid )
	    return;
	requireUniqueInstance();
        for ( int i = 0; i < cacheHits.size(); i++ ) {
            DiscDO DO = ( DiscDO ) cacheHits.elementAt( i );
            if ( null == DO ) continue;
            boolean equals = true;
	    ObjectId id = DO.getOId();
	    if ( null == id || ! id.equals( oid ) )
		cacheHits.removeElementAt( i-- );
	}

	// Also prepare the SQL needed to query the database
        // in case there is no cache, or the query involves other tables.
	builder.addWhere( DiscDO.PrimaryKey, oid.toBigDecimal(),
				QueryBuilder.EQUAL );
    }

    /**
     * Set the object handle to query.
     * This is a variant of setQueryOId().
     *
     * @param handle The string version of the id to query.
     */
    public void setQueryHandle(String handle)
    throws ObjectIdException {
        ObjectId oid = new ObjectId( handle );
        setQueryOId( oid );
    }


    /**
     * Set "unique instance" assertion bit.
     * The first call to the next() method will throw an exception
     * if more than one object was found.
     */
    public void requireUniqueInstance()
    {
	uniqueInstance = true;
    }


    /**
     * Prepare this DiscQuery object to be used for another query.
     */
    public void reset() {
	cacheHits	= new Vector();
	DOs		= null;
	uniqueInstance	= false;
	needToRun	= true;
	doneResultSet	= true;

        builder.reset();

    }


    /**
     * Release the DBQuery object.
     */
    public void release() {

	if ( null != dbQuery ) {
	    dbQuery.release();
	    dbQuery = null;
	}

    }

    //
    // Implementation of Query interface
    //

    /**
     * Run the query against the database, and return the ResultSet.
     *
     * The following call in runQuery()
     *	    dbQuery.query( this );
     * causes the dbQuery object to invoke
     *      executeQuery()
     *
     * @param conn Handle to database connection.
     * @exception java.sql.SQLException If a database access error occurs.
     */
    public ResultSet executeQuery(DBConnection conn)
        throws SQLException
    {
	resultSet = builder.executeQuery( conn );
        return resultSet;
    }
    public ResultSet executeQuery(Connection conn)
        throws SQLException
    {
	resultSet = builder.executeQuery( conn );
        return resultSet;
    }


    /**
     * WARNING!  This method is disabled.
     * It's implementation is forced by the Query interface.
     * This method is disabled for 2 reasons:
     * 1)  the getDOArray() and getNextDO() methods are better
     *     because they return DOs instead of JDBC objects.
     * 2)  the createExisting() method throws an exception
     *     that we cannot reasonably handle here,
     *     and that we cannot throw from here.
     *
     * @param rs JDBC result set from which the next object
     *   will be instantiated.
     * @exception java.sql.SQLException
     *   If a database access error occurs.
     * @exception com.lutris.appserver.server.sql.ObjectIdException
     *   If an invalid object id was queried from the database.
     */
    public Object next(ResultSet rs)
    throws SQLException, ObjectIdException {
	// TODO: It would be nice to throw an unchecked exception here
	// (an exception that extends RuntimeException)
	// that would be guaranteed to appear during application testing.
	throw new ObjectIdException(
	    "next() should not be used.  Use getNextDO() instead." );
	//return null;
    }




    /**
     * Only used with executeUpdate().
     * Prepare to update column title in table Disc
     * for all rows would be returned by getDOArray().
     *
     * @param x The title of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
   public void setUpdateTitle( String x )
   throws QueryException
   // DataObjectException, 
    {
        try {
	    builder.addUpdate( DiscDO.Title, x );
        } catch ( QueryException e ) { 
            throw new QueryException( 
		"DiscQuery.setUpdateTitle(): ", e );
        }
    }

    /**
     * Set the title to query, with a QueryBuilder comparison operator.
     *
     * @param x The title of the Disc to query.
     * @param cmp_op QueryBuilder comparison operator to use.
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryTitle(
				String x,
				String cmp_op )
    throws DataObjectException, QueryException
    {
	// Remove from cacheHits any DOs that do not meet this
	// setQuery requirement.
	for ( int i = 0; i < cacheHits.size() && ! hitDb; i++ ) {
	    DiscDO DO = ( DiscDO ) cacheHits.elementAt( i );
	    if ( null == DO ) continue;
	    boolean found = false;
	    String m = DO.getTitle();
	    found = QueryBuilder.compare( m, x, cmp_op );
	    if ( ! found )
		cacheHits.removeElementAt( i-- );
	}

	// Also prepare the SQL needed to query the database 
	// in case there is no cache, or the query involves other tables.
	if ( partialCache || hitDb )
	    builder.addWhere( DiscDO.Title, x, cmp_op );
    }

    /**
     * Set the title to query.
     *
     * @param x The title of the Disc to query.
     * @param exact to use matches or not
     * @exception DataObjectException If a database access error occurs.
     * @deprecated Instead use 
     * setQueryTitle(String x,String cmp_op)
     */
    public void setQueryTitle(
				String x, boolean exact)
    throws DataObjectException, QueryException
    {
	// TODO:  DELETE THIS METHOD IN A FUTURE RELEASE
	String cmp_op = QueryBuilder.EQUAL;
	if ( ! exact )
	    cmp_op = QueryBuilder.CASE_INSENSITIVE_CONTAINS;
	setQueryTitle( x, cmp_op );
    }

    /**
     * Set the title to query
     * @param x The title of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
    public void setQueryTitle( 
				String x )
    throws DataObjectException, QueryException {
	setQueryTitle( x, true );
    }

    /**
     * Add title to the ORDER BY clause.
     *
     * @param direction_flag  True for ascending order, false for descending
     */
    public void addOrderByTitle(boolean direction_flag) {
	hitDb = true;
        builder.addOrderByColumn("title",
					(direction_flag) ? "ASC" : "DESC");
    }


    /**
     * Add title to the ORDER BY clause.  This convenience
     * method assumes ascending order.
     */
    public void addOrderByTitle() {
        builder.addOrderByColumn("title","ASC");
    }




    /**
     * Only used with executeUpdate().
     * Prepare to update column artist in table Disc
     * for all rows would be returned by getDOArray().
     *
     * @param x The artist of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
   public void setUpdateArtist( String x )
   throws QueryException
   // DataObjectException, 
    {
        try {
	    builder.addUpdate( DiscDO.Artist, x );
        } catch ( QueryException e ) { 
            throw new QueryException( 
		"DiscQuery.setUpdateArtist(): ", e );
        }
    }

    /**
     * Set the artist to query, with a QueryBuilder comparison operator.
     *
     * @param x The artist of the Disc to query.
     * @param cmp_op QueryBuilder comparison operator to use.
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryArtist(
				String x,
				String cmp_op )
    throws DataObjectException, QueryException
    {
	// Remove from cacheHits any DOs that do not meet this
	// setQuery requirement.
	for ( int i = 0; i < cacheHits.size() && ! hitDb; i++ ) {
	    DiscDO DO = ( DiscDO ) cacheHits.elementAt( i );
	    if ( null == DO ) continue;
	    boolean found = false;
	    String m = DO.getArtist();
	    found = QueryBuilder.compare( m, x, cmp_op );
	    if ( ! found )
		cacheHits.removeElementAt( i-- );
	}

	// Also prepare the SQL needed to query the database 
	// in case there is no cache, or the query involves other tables.
	if ( partialCache || hitDb )
	    builder.addWhere( DiscDO.Artist, x, cmp_op );
    }

    /**
     * Set the artist to query.
     *
     * @param x The artist of the Disc to query.
     * @param exact to use matches or not
     * @exception DataObjectException If a database access error occurs.
     * @deprecated Instead use 
     * setQueryArtist(String x,String cmp_op)
     */
    public void setQueryArtist(
				String x, boolean exact)
    throws DataObjectException, QueryException
    {
	// TODO:  DELETE THIS METHOD IN A FUTURE RELEASE
	String cmp_op = QueryBuilder.EQUAL;
	if ( ! exact )
	    cmp_op = QueryBuilder.CASE_INSENSITIVE_CONTAINS;
	setQueryArtist( x, cmp_op );
    }

    /**
     * Set the artist to query
     * @param x The artist of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
    public void setQueryArtist( 
				String x )
    throws DataObjectException, QueryException {
	setQueryArtist( x, true );
    }

    /**
     * Add artist to the ORDER BY clause.
     *
     * @param direction_flag  True for ascending order, false for descending
     */
    public void addOrderByArtist(boolean direction_flag) {
	hitDb = true;
        builder.addOrderByColumn("artist",
					(direction_flag) ? "ASC" : "DESC");
    }


    /**
     * Add artist to the ORDER BY clause.  This convenience
     * method assumes ascending order.
     */
    public void addOrderByArtist() {
        builder.addOrderByColumn("artist","ASC");
    }




    /**
     * Only used with executeUpdate().
     * Prepare to update column genre in table Disc
     * for all rows would be returned by getDOArray().
     *
     * @param x The genre of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
   public void setUpdateGenre( String x )
   throws QueryException
   // DataObjectException, 
    {
        try {
	    builder.addUpdate( DiscDO.Genre, x );
        } catch ( QueryException e ) { 
            throw new QueryException( 
		"DiscQuery.setUpdateGenre(): ", e );
        }
    }

    /**
     * Set the genre to query, with a QueryBuilder comparison operator.
     *
     * @param x The genre of the Disc to query.
     * @param cmp_op QueryBuilder comparison operator to use.
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryGenre(
				String x,
				String cmp_op )
    throws DataObjectException, QueryException
    {
	// Remove from cacheHits any DOs that do not meet this
	// setQuery requirement.
	for ( int i = 0; i < cacheHits.size() && ! hitDb; i++ ) {
	    DiscDO DO = ( DiscDO ) cacheHits.elementAt( i );
	    if ( null == DO ) continue;
	    boolean found = false;
	    String m = DO.getGenre();
	    found = QueryBuilder.compare( m, x, cmp_op );
	    if ( ! found )
		cacheHits.removeElementAt( i-- );
	}

	// Also prepare the SQL needed to query the database 
	// in case there is no cache, or the query involves other tables.
	if ( partialCache || hitDb )
	    builder.addWhere( DiscDO.Genre, x, cmp_op );
    }

    /**
     * Set the genre to query.
     *
     * @param x The genre of the Disc to query.
     * @param exact to use matches or not
     * @exception DataObjectException If a database access error occurs.
     * @deprecated Instead use 
     * setQueryGenre(String x,String cmp_op)
     */
    public void setQueryGenre(
				String x, boolean exact)
    throws DataObjectException, QueryException
    {
	// TODO:  DELETE THIS METHOD IN A FUTURE RELEASE
	String cmp_op = QueryBuilder.EQUAL;
	if ( ! exact )
	    cmp_op = QueryBuilder.CASE_INSENSITIVE_CONTAINS;
	setQueryGenre( x, cmp_op );
    }

    /**
     * Set the genre to query
     * @param x The genre of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
    public void setQueryGenre( 
				String x )
    throws DataObjectException, QueryException {
	setQueryGenre( x, true );
    }

    /**
     * Add genre to the ORDER BY clause.
     *
     * @param direction_flag  True for ascending order, false for descending
     */
    public void addOrderByGenre(boolean direction_flag) {
	hitDb = true;
        builder.addOrderByColumn("genre",
					(direction_flag) ? "ASC" : "DESC");
    }


    /**
     * Add genre to the ORDER BY clause.  This convenience
     * method assumes ascending order.
     */
    public void addOrderByGenre() {
        builder.addOrderByColumn("genre","ASC");
    }




    /**
     * Only used with executeUpdate().
     * Prepare to update column owner in table Disc
     * for all rows would be returned by getDOArray().
     *
     * @param x The owner of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
   public void setUpdateOwner( org.enhydra.barracuda.discRack.data.person.PersonDO x )
   throws QueryException
   // DataObjectException, 
    {
        try {
	    builder.addUpdate( DiscDO.Owner, x );
        } catch ( QueryException e ) { 
            throw new QueryException( 
		"DiscQuery.setUpdateOwner(): ", e );
        }
    }

    /**
     * Set the owner to query, with a QueryBuilder comparison operator.
     *
     * @param x The owner of the Disc to query.
     * @param cmp_op QueryBuilder comparison operator to use.
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryOwner(
				org.enhydra.barracuda.discRack.data.person.PersonDO x,
				String cmp_op )
    throws DataObjectException, QueryException
    {
	// Remove from cacheHits any DOs that do not meet this
	// setQuery requirement.
	for ( int i = 0; i < cacheHits.size() && ! hitDb; i++ ) {
	    DiscDO DO = ( DiscDO ) cacheHits.elementAt( i );
	    if ( null == DO ) continue;
	    boolean found = false;
	    org.enhydra.barracuda.discRack.data.person.PersonDO m = DO.getOwner();
	    found = QueryBuilder.compare( m, x, cmp_op );
	    if ( ! found )
		cacheHits.removeElementAt( i-- );
	}

	// Also prepare the SQL needed to query the database 
	// in case there is no cache, or the query involves other tables.
	if ( partialCache || hitDb )
	    builder.addWhere( DiscDO.Owner, x, cmp_op );
    }

    /**
     * Set the owner to query.
     *
     * @param x The owner of the Disc to query.
     * @param exact to use matches or not
     * @exception DataObjectException If a database access error occurs.
     * @deprecated Instead use 
     * setQueryOwner(org.enhydra.barracuda.discRack.data.person.PersonDO x,String cmp_op)
     */
    public void setQueryOwner(
				org.enhydra.barracuda.discRack.data.person.PersonDO x, boolean exact)
    throws DataObjectException, QueryException
    {
	// TODO:  DELETE THIS METHOD IN A FUTURE RELEASE
	String cmp_op = QueryBuilder.EQUAL;
	if ( ! exact )
	    cmp_op = QueryBuilder.CASE_INSENSITIVE_CONTAINS;
	setQueryOwner( x, cmp_op );
    }

    /**
     * Set the owner to query
     * @param x The owner of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
    public void setQueryOwner( 
				org.enhydra.barracuda.discRack.data.person.PersonDO x )
    throws DataObjectException, QueryException {
	setQueryOwner( x, true );
    }

    /**
     * Add owner to the ORDER BY clause.
     *
     * @param direction_flag  True for ascending order, false for descending
     */
    public void addOrderByOwner(boolean direction_flag) {
	hitDb = true;
        builder.addOrderByColumn("owner",
					(direction_flag) ? "ASC" : "DESC");
    }


    /**
     * Add owner to the ORDER BY clause.  This convenience
     * method assumes ascending order.
     */
    public void addOrderByOwner() {
        builder.addOrderByColumn("owner","ASC");
    }




    /**
     * Only used with executeUpdate().
     * Prepare to update column isLiked in table Disc
     * for all rows would be returned by getDOArray().
     *
     * @param x The isLiked of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
   public void setUpdateIsLiked( boolean x )
   throws QueryException
   // DataObjectException, 
    {
        try {
	    builder.addUpdate( DiscDO.IsLiked, x );
        } catch ( QueryException e ) { 
            throw new QueryException( 
		"DiscQuery.setUpdateIsLiked(): ", e );
        }
    }

    /**
     * Set the isLiked to query, with a QueryBuilder comparison operator.
     *
     * @param x The isLiked of the Disc to query.
     * @param cmp_op QueryBuilder comparison operator to use.
     * @exception DataObjectException If a database access error occurs.
     * @exception QueryException If comparison operator is inappropriate
     * (e.g. CASE_SENSITIVE_CONTAINS on an integer field).
     */
    public void setQueryIsLiked(
				boolean x,
				String cmp_op )
    throws DataObjectException, QueryException
    {
	// Remove from cacheHits any DOs that do not meet this
	// setQuery requirement.
	for ( int i = 0; i < cacheHits.size() && ! hitDb; i++ ) {
	    DiscDO DO = ( DiscDO ) cacheHits.elementAt( i );
	    if ( null == DO ) continue;
	    boolean found = false;
	    boolean m = DO.getIsLiked();
	    found = QueryBuilder.compare( m, x, cmp_op );
	    if ( ! found )
		cacheHits.removeElementAt( i-- );
	}

	// Also prepare the SQL needed to query the database 
	// in case there is no cache, or the query involves other tables.
	if ( partialCache || hitDb )
	    builder.addWhere( DiscDO.IsLiked, x, cmp_op );
    }

    /**
     * Set the isLiked to query.
     *
     * @param x The isLiked of the Disc to query.
     * @param exact to use matches or not
     * @exception DataObjectException If a database access error occurs.
     * @deprecated Instead use 
     * setQueryIsLiked(boolean x,String cmp_op)
     */
    public void setQueryIsLiked(
				boolean x, boolean exact)
    throws DataObjectException, QueryException
    {
	// TODO:  DELETE THIS METHOD IN A FUTURE RELEASE
	String cmp_op = QueryBuilder.EQUAL;
	if ( ! exact )
	    cmp_op = QueryBuilder.CASE_INSENSITIVE_CONTAINS;
	setQueryIsLiked( x, cmp_op );
    }

    /**
     * Set the isLiked to query
     * @param x The isLiked of the Disc to query.
     * @exception DataObjectException If a database access error occurs.
     */
    public void setQueryIsLiked( 
				boolean x )
    throws DataObjectException, QueryException {
	setQueryIsLiked( x, true );
    }

    /**
     * Add isLiked to the ORDER BY clause.
     *
     * @param direction_flag  True for ascending order, false for descending
     */
    public void addOrderByIsLiked(boolean direction_flag) {
	hitDb = true;
        builder.addOrderByColumn("isLiked",
					(direction_flag) ? "ASC" : "DESC");
    }


    /**
     * Add isLiked to the ORDER BY clause.  This convenience
     * method assumes ascending order.
     */
    public void addOrderByIsLiked() {
        builder.addOrderByColumn("isLiked","ASC");
    }


    

    /**
    * Returns the <code>QueryBuilder</code> that this <code>DiscQuery</code>
    * uses to construct and execute database queries.
    * <code>DiscQuery.setQueryXXX</code> methods use 
    * the <code>QueryBuilder</code> to
    * append SQL expressions to the <code>"WHERE"</code> clause to be executed.
    * The <code>QueryBuilder.addEndClause method.</code> can be used to
    * append freeform SQL expressions to the <code>WHERE</code> clause,
    * e.g. "ORDER BY name".
    *
    * <b>Notes regarding cache-enabled DO classes:</b>
    * DO classes can be cache-enabled.  
    * If when using a <code>DiscQuery</code>, the application developer
    * <b>does not call</b> <code>getQueryBuilder</code>,
    * then <code>DiscQuery.setQueryXXX</code> methods 
    * simply prune the DO cache and return the remaining results.
    * However, a <code>QueryBuilder</code> builds
    * <CODE>SELECT</CODE> statements for execution by the actual database,
    * and never searches the built-in cache for the DO.
    * So, if the DO class is cache-enabled, and <code>getQueryBuilder</code>
    * is called, this <CODE>DiscQuery</CODE> object ignores the cache 
    * and hits the actual database.
    */
    public QueryBuilder getQueryBuilder() {
	hitDatabase();
	return builder;
    }

    /**
     * Insert an <CODE>OR</CODE> between <CODE>WHERE</CODE> clauses.
     * Example:  find all the persons named Bob or Robert:
     * <CODE>
     *    PersonQuery pq = new PersonQuery();
     *    pq.setQueryFirstName( "Bob" );
     *    pq.or();
     *    pq.setQueryFirstName( "Robert" );
     * </CODE>
     * 
     * Note:  Calls to <CODE>setQueryXxx</CODE> methods
     * are implicitly <CODE>AND</CODE>ed together,
     * so the following example will always return nothing:
     * <CODE>
     *    PersonQuery pq = new PersonQuery();
     *    pq.setQueryFirstName( "Bob" );
     *    // AND automatically inserted here.
     *    pq.setQueryFirstName( "Robert" );
     * </CODE>
     * 
     * @see QueryBuilder to construct more elaborate queries.
     * @author Jay Gunter
     */
    public void or() {
	hitDb = true;
	builder.addWhereOr();
    }

    /**
     * Place an open parenthesis in the <CODE>WHERE</CODE> clause.
     * Example usage:  find all the Bobs and Roberts who are 5 or 50 years old:
     * <CODE>
     *    PersonQuery pq = new PersonQuery();
     *    pq.openParen();
     *       pq.setQueryFirstName( "Bob" );
     *       pq.or();
     *       pq.setQueryFirstName( "Robert" );
     *    pq.closeParen();
     *    // AND automatically inserted here.
     *    pq.openParen();
     *       pq.setQueryAge( 5 );
     *       pq.or();
     *       pq.setQueryAge( 50 );
     *    pq.closeParen();
     * </CODE>
     * 
     * @see QueryBuilder to construct more elaborate queries.
     * @author Jay Gunter
     */
    public void openParen() {
        hitDb = true;
	builder.addWhereOpenParen();
    }

    /**
     * Place a closing parenthesis in the <CODE>WHERE</CODE> clause.
     * 
     * @see openParen
     * @author Jay Gunter
     */
    public void closeParen() {
        hitDb = true;
	builder.addWhereCloseParen();
    }


}
