/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: DiscScreen.java,v 1.15 2003/02/03 21:47:47 jacobk Exp $
 */
package org.enhydra.barracuda.discRack.pres.screens;

import java.io.*;
import java.util.*;
import java.net.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.w3c.dom.*;
import org.w3c.dom.html.*;

import com.lutris.appserver.server.Enhydra;
import com.lutris.logging.LogChannel;

import org.enhydra.xml.xmlc.*;

import org.enhydra.barracuda.core.comp.*;
import org.enhydra.barracuda.core.event.*;
import org.enhydra.barracuda.core.event.helper.*;
import org.enhydra.barracuda.core.forms.*;
import org.enhydra.barracuda.core.forms.validators.*;
import org.enhydra.barracuda.plankton.data.*;
import org.enhydra.barracuda.core.util.dom.*;
import org.enhydra.barracuda.core.util.http.*;
import org.apache.log4j.*;
import org.enhydra.barracuda.core.util.srv.*;

import org.enhydra.barracuda.discRack.biz.*;
import org.enhydra.barracuda.discRack.biz.person.*;
import org.enhydra.barracuda.discRack.biz.disc.*;
import org.enhydra.barracuda.discRack.pres.events.*;
import org.enhydra.barracuda.discRack.pres.services.*;
import org.enhydra.barracuda.discRack.pres.xmlc.*;

/**
 * <p>Event handlers (both Controller and View) for the
 * Disc screen. Note that we actually handle both the Catalog
 * screen here AND the Disc Maintenance screens.
 *
 * <p>We handle the following Control events:
 * <ul>
 *		<li>GetCatalog - uses an EventForwardingFactory to
 *			automatically fire a RenderCatalog event</li>
 *		<li>GetCatalogMaint - uses an EventForwardingFactory to
 *			automatically fire a RenderCatalogMaint event</li>
 *		<li>GetCatalogEdit - create the disc form, map the request to the
 *			form and then validate it. If there weren't any errors,
 *			then it creates an edit form and populates it from the
 *			currently selected disc record, and saves the edit form in
 *			the event context so that when the RenderCatalogMaint event
 *			is handled the screen can be populated with data. If the
 *			disc form is invalid (ie. user didn't select a disc) then
 *			fire a GetCatalog event.
 *		<li>DoCatalogMaint - this event handler handles both DoCatalogUpdate
 *			and DoCatalogDelete events. Instantiates the Edit form, figures out
 *			what the appropriate action should be (update or delete), does the
 *			deed, and then redirects to the GetCatalog or GetCatalogMaint event,
 *			depending on whether or not there were errors</li>
 * </ul>
 *
 * <p>We handle the following View events:
 * <ul>
 *		<li>RenderCatalog - Actually render the catalog screen</li>
 *		<li>RenderCatalogMaint - Actually render the edit screen. If the
 *			event context contains an EditForm, we repopulate the screen
 *			from the form. If the context contains edit errs, we will show
 *			those as well</li>
 * </ul>
 *
 * <p>We define the following Forms:
 * <ul>
 *		<li>DiscForm - contains only one form element: ID. This form
 *			is used primarily to ensure that the user selected a record
 *			on the catalog screen prior to pressing update or delete.</li>
 *		<li>EditForm - contains 5 elements: ID, ARTIST, TITLE, GENRE, and
 *			IS_LIKED, all of which are Strings and take NotNullValidators
 *			_except_ for IS_LIKED, which is a Boolean. In addition, the
 *			edit form uses an IDValidator to ensure that the ID is valid for
 *			given actions.</li>
 * </ul>
 *
 * <p>You can <a href="sm_barracudaDiscRack_disc.gif">click here to see a diagram...</a>
 */
public class DiscScreen extends DefaultEventGateway {

	//public constants
	protected static Logger logger = Logger.getLogger(DiscScreen.class.getName());

	//disc actions
	private static final int ADD = 0;
	private static final int CHANGE = 1;
	private static final int DELETE = 2;

	//form map constants
	private static final String ACTION = "Action";								//Integer
	private static final String DISC = "Disc";									//Disc

	//private constants
	private static final String DISC_ERR = DiscScreen.class.getName()+".DiscErr";			//ValidationException
	private static final String EDIT_FORM = DiscScreen.class.getName()+".EditForm";			//EditForm
	private static final String EDIT_ERR = DiscScreen.class.getName()+".EditErr";			//ValidationException

	//this defines the various event handlers
	private ListenerFactory getCatalogFactory = 		new EventForwardingFactory(new RenderCatalog());
	private ListenerFactory getCatalogEditFactory = 	new DefaultListenerFactory() {public BaseEventListener getInstance() {return new GetCatalogEditHandler();}		public String getListenerID() {return getID(GetCatalogEditHandler.class);}};
	private ListenerFactory getCatalogMaintFactory = 	new EventForwardingFactory(new RenderCatalogMaint());
	private ListenerFactory doCatalogMaintFactory = 	new DefaultListenerFactory() {public BaseEventListener getInstance() {return new DoCatalogMaintHandler();}		public String getListenerID() {return getID(DoCatalogMaintHandler.class);}};
	private ListenerFactory renderCatalogFactory = 		new DefaultListenerFactory() {public BaseEventListener getInstance() {return new RenderCatalogHandler();}		public String getListenerID() {return getID(RenderCatalogHandler.class);}};
	private ListenerFactory renderCatalogMaintFactory =	new DefaultListenerFactory() {public BaseEventListener getInstance() {return new RenderCatalogMaintHandler();}	public String getListenerID() {return getID(RenderCatalogMaintHandler.class);}};
	//private vars


	//-------------------- DefaultEventGateway -------------------
	/**
	 * Public constructor
	 */
	public DiscScreen() {
		//specify who's interested in what
		specifyLocalEventInterests(getCatalogFactory, GetCatalog.class);
		specifyLocalEventInterests(getCatalogEditFactory, GetCatalogEdit.class);
		specifyLocalEventInterests(getCatalogMaintFactory, GetCatalogMaint.class);
		specifyLocalEventInterests(doCatalogMaintFactory, DoCatalogMaint.class);
		specifyLocalEventInterests(renderCatalogFactory, RenderCatalog.class);
		specifyLocalEventInterests(renderCatalogMaintFactory, RenderCatalogMaint.class);

		//specify aliases for events we're not explicitly registering an interest in
		specifyLocalEventAliases(GetCatalogAdd.class);
		specifyLocalEventAliases(DoCatalogUpdate.class);
		specifyLocalEventAliases(DoCatalogDelete.class);
	}


	//------------------------------------------------------------
	//             Model 2 - Controller Event Handlers
	//------------------------------------------------------------
	/**
	 * GetCatalogEditHandler - this is where we handle a request to
	 * edit the disc record
	 */
	class GetCatalogEditHandler extends DefaultBaseEventListener {
		public void handleControlEvent(ControlEventContext context) throws EventException, ServletException, IOException {
			//unpack the necessary entities from the context
			HttpServletRequest req = context.getRequest();

			//start by creating the form
			if (logger.isDebugEnabled()) logger.debug("Creating id form");
			DiscForm df = new DiscForm();
			EditForm ef = null;

			try {
				//map, validate the id form
				if (logger.isDebugEnabled()) logger.debug("Mapping id form");

				df.map(req).validate(true);
				String id = df.getStringVal(DiscForm.ID);

				//get the Disc (the validator will have looked this up)
				Disc disc = (Disc) df.getState(DISC);

				//now build the edit map by creating a statemap from the disc record
				//and then mapping the edit form from that
				if (logger.isDebugEnabled()) logger.debug("Mapping disc record to disc form");
				if (disc!=null) {
					StateMap sm = new DefaultStateMap();
					try {sm.putState(EditForm.ID, disc.getHandle());} catch (DiscRackBusinessException e) {}
					try {sm.putState(EditForm.ARTIST, disc.getArtist());} catch (DiscRackBusinessException e) {}
					try {sm.putState(EditForm.TITLE, disc.getTitle());} catch (DiscRackBusinessException e) {}
					try {sm.putState(EditForm.GENRE, disc.getGenre());} catch (DiscRackBusinessException e) {}
					try {sm.putState(EditForm.IS_LIKED, new Boolean(disc.isLiked()));} catch (DiscRackBusinessException e) {}
					ef = new EditForm();
					ef.map(sm);
				}
			} catch (ValidationException e) {
				if (logger.isDebugEnabled()) logger.debug("Validation Exception: "+e);
				if (logger.isDebugEnabled()) CollectionsUtil.printStackTrace(e.getExceptionList(), 0, logger, null);
				context.putState(DISC_ERR, e);
				throw new InterruptDispatchException(new GetCatalog());
			}

			//save it in the queue for rendering
			if (logger.isDebugEnabled()) logger.debug("Saving form, errors");
			context.putState(EDIT_FORM, ef);
			context.putState(EDIT_ERR, null);
		}
	}

	/**
	 * DoCatalogMaintHandler - this is where we handle any attempt to
	 * actually update the disc
	 */
	class DoCatalogMaintHandler extends DefaultBaseEventListener {
		public void handleControlEvent(ControlEventContext context) throws EventException, ServletException, IOException {
			//unpack the necessary entities from the context
			BaseEvent event = context.getEvent();
			HttpServletRequest req = context.getRequest();

			//start by creating the form
			if (logger.isDebugEnabled()) logger.debug("Creating edit form");
			EditForm ef = new EditForm();

			ValidationException ve = null;
			int action = -1;
			try {
				//map the form
				if (logger.isDebugEnabled()) logger.debug("Mapping edit form");
				ef.map(req);
				String id = ef.getStringVal(EditForm.ID);

				//figure out what the underlying event actually was to
				//determine what "mode" (add, change, delete) we're in.
				//store this value in the form map (the validator will use it)
				if (logger.isDebugEnabled()) logger.debug("Determining action: ");
				BaseEvent rootEvent = DefaultBaseEvent.getOriginalEvent(event);
				if (rootEvent instanceof DoCatalogUpdate) {
					if (id.equals(EditForm.INVALID_ID)) action = ADD;
					else action = CHANGE;
				} else if (rootEvent instanceof DoCatalogDelete) {
					action = DELETE;
				}
				ef.putState(ACTION, new Integer(action));

				//validate the form. If we're in delete mode, just validate the form
				//otherwise, validate the whole thing
				if (logger.isDebugEnabled()) logger.debug("Validating edit form (action="+action+")");
				if (action==DELETE) ef.validateForm(true);
				else ef.validate(true);

				//get the Disc (the validator will have looked this up)
				Disc disc = (Disc) ef.getState(DISC);

				//update based on action
				try {
					if (action==DELETE) {
						if (logger.isDebugEnabled()) logger.debug("Deleting Disc");
						if (disc!=null) disc.delete();
					} else {
						if (logger.isDebugEnabled()) logger.debug("Updating Disc");
						if (disc==null) disc = new Disc();
				        disc.setArtist(ef.getStringVal(EditForm.ARTIST));
				        disc.setTitle(ef.getStringVal(EditForm.TITLE));
				        disc.setGenre(ef.getStringVal(EditForm.GENRE));
				        disc.setLiked(ef.getBooleanVal(EditForm.IS_LIKED).booleanValue());
				        disc.setOwner(LoginServices.getCurrentUser(SessionServices.getSession(req)));
				        disc.save();
					}
			    } catch (DiscRackBusinessException e) {
					throw new ValidationException(this, "Error updating disc: "+e.getMessage()+" Please try again. If the problem persists contact your system administrator.", e);
			    }
			} catch (ValidationException e) {
				if (logger.isDebugEnabled()) logger.debug("Validation Exception: "+e);
				if (logger.isDebugEnabled()) CollectionsUtil.printStackTrace(e.getExceptionList(), 0, logger, null);
				ve = e;
			}

			//store a copy of the form and any errors in the queue
			if (logger.isDebugEnabled()) logger.debug("Saving form, errors");
			context.putState(EDIT_FORM, ef);
			context.putState(EDIT_ERR, ve);

			//redirect appropriately: if it's valid, update the session
			//and head for the catalog screen. Otherwise, return to the
			//edit screen...
			if (ve==null) {
				if (logger.isDebugEnabled()) logger.debug("Redirect to GetCatalog");
				throw new ClientSideRedirectException(new GetCatalog());
			} else if (action==DELETE) {
				context.putState(DISC_ERR, ve);
				if (logger.isDebugEnabled()) logger.debug("Redirect Delete to GetCatalog");
				throw new ClientSideRedirectException(new GetCatalog());
			} else {
				if (logger.isDebugEnabled()) logger.debug("Redirect to GetCatalogMaint");
				throw new ClientSideRedirectException(new GetCatalogMaint());
			}
	    }
	}


	//------------------------------------------------------------
	//                Model 2 - View Event Handlers
	//------------------------------------------------------------
	/**
	 * RenderCatalogHandler -
	 */
	class RenderCatalogHandler extends DefaultBaseEventListener {
		public void handleViewEvent(ViewEventContext context) throws EventException, ServletException, IOException {
			//unpack the necessary entities from the context
			HttpServletRequest req = context.getRequest();

			//get the XMLC object
//			XMLCFactory factory = XMLCServices.getXMLCFactory(context.getConfig());
//			DiscCatalogHTML page = (DiscCatalogHTML) factory.create(DiscCatalogHTML.class);
            DiscCatalogHTML page = (DiscCatalogHTML) DefaultDOMLoader.getGlobalInstance().getDOM(DiscCatalogHTML.class);


//			// Swap in our real run-time JavaScript to replace the storyboard JavaScript
//			HTMLScriptElement script = new DiscCatalogScriptHTML().getElementRealScript();
//			XMLCUtil.replaceNode(script, page.getElementDummyScript());

/*
			//find out who we are logged in as
			if (logger.isDebugEnabled()) logger.debug("Looking up Person record");
			Person person = LoginServices.getCurrentUser(SessionServices.getSession(req));
		    try {
				page.setTextOwner(person.getFirstname()+" "+person.getLastname()+"'s Discs");
		    } catch (DiscRackBusinessException e) {
				page.setTextOwner("Error getting user info: "+e);
				Debug.println(this, 0, "Error getting user info: "+e);
				Debug.printStackTrace(e);
		    }

			//check for errors
			if (logger.isDebugEnabled()) logger.debug("Checking for errors");
			ValidationException ve = (ValidationException) context.get(DISC_ERR);
			if (ve==null) {
				if (showDebug>2) Debug.println (this, 2, "No errors found");
			    page.getElementErrorText().getParentNode().removeChild(page.getElementErrorText());
			} else {
				List errlist = ve.getExceptionList();
				if (showDebug>2) Debug.println (this, 2, errlist.size()+"error(s) found");
				StringBuffer sb = new StringBuffer(errlist.size()>1 ? "There were several errors:" : "");
				Iterator it = errlist.iterator();
				while (it.hasNext()) {
					sb.append(" "+((ValidationException) it.next()).getMessage());
				}
			    page.setTextErrorText(sb.toString());
			}

			// Generate the person's Disc list and create the HTML template references
			HTMLTableRowElement templateRow = page.getElementTemplateRow();
			HTMLOptionElement templateOption = page.getElementTemplateOption();
			Node discTable = templateRow.getParentNode();
			Node discSelect = templateOption.getParentNode();

			// Remove ids to prevent duplicates
			templateRow.removeAttribute("id");
			templateOption.removeAttribute("id");

			// Remove id attributes from the cells in the template row
			HTMLElement artistCellTemplate = page.getElementArtist();
			HTMLElement titleCellTemplate = page.getElementTitle();
			HTMLElement genreCellTemplate = page.getElementGenre();
			HTMLElement likeThisDisc = page.getElementLikeThisDisc();

			artistCellTemplate.removeAttribute("id");
			titleCellTemplate.removeAttribute("id");
			genreCellTemplate.removeAttribute("id");
			likeThisDisc.removeAttribute("id");

			// Remove the dummy storyboard text from the prototype HTML
			templateOption.removeChild(templateOption.getFirstChild());

			try {
			    Disc[] discList = DiscFactory.findDiscsForPerson(person);

			    // Get collection of Discs and loop through collection
			    // to add each disc as a row in the table.
			    for (int numDiscs=0; numDiscs<discList.length; numDiscs++) {

			        // set text of new cells to values from string array
			        Disc currentDisc = discList[numDiscs];
			        page.setTextArtist(currentDisc.getArtist());
			        page.setTextTitle(currentDisc.getTitle());
			        page.setTextGenre(currentDisc.getGenre());
		            page.setTextLikeThisDisc(currentDisc.isLiked() ? "Yes" : "No");

			    	// Add a deep clone of the row to the DOM
			        Node clonedNode = templateRow.cloneNode(true);
			        discTable.appendChild(clonedNode);
			        // Alternative way to insert nodes below
			        // insertBefore(newNode, oldNode);
			        // discTable.insertBefore(clonedNode, templateRow);

			        // Now populate the select list
			        // This algorithm is obscure because options are not normal
			        //  HTML elements
			        // First populate the option value (the disc database ID).
			        //  Then append a text child as the option
			        // text, which is what is displayed as the text
			        // in each row of the select box
			        HTMLOptionElement clonedOption = (HTMLOptionElement) templateOption.cloneNode(true);
			        clonedOption.setValue(currentDisc.getHandle());
			        Node optionTextNode = clonedOption.getOwnerDocument().createTextNode(currentDisc.getArtist()+": "+currentDisc.getTitle());
			        clonedOption.appendChild(optionTextNode);
			        // Do only a shallow copy of the option as we don't want the text child
			        // of the node option
			        discSelect.appendChild(clonedOption);
			        // Alternative way to insert nodes below
			        // insertBefore(newNode, oldNode);
			        // discSelect.insertBefore(clonedOption, templateOption);
			    }
			} catch(Exception e) {
				Debug.println(this, 1, "Error populating disc catalog: "+e);
				Debug.printStackTrace(e);
			}

			// Finally remove the template row and template select option
			//  from the page
			templateRow.getParentNode().removeChild(templateRow);
			templateOption.getParentNode().removeChild(templateOption);
*/
			//load the person, discs list info (we do this here since both models
			//created below need the same info; might as well only load it once)
			if (logger.isDebugEnabled()) logger.debug("Loading person, disc info");
			Person person = LoginServices.getCurrentUser(SessionServices.getSession(context.getRequest()));
			Disc[] discs = null;
		    try {discs = DiscFactory.findDiscsForPerson(person);}
			catch (DiscRackBusinessException e) {if (logger.isDebugEnabled()) logger.debug("Err getting disc list:"+e);}
			ValidationException ve = (ValidationException) context.getState(DISC_ERR);

			//create a template component and render it
			if (logger.isDebugEnabled()) logger.debug("Creating template component");
			Node node = page.getDocument().getElementById("CatalogForm");
			BTemplate templateComp = new BTemplate();
            templateComp.setView(new DefaultTemplateView(node));
			templateComp.addModel(new CatalogModel(person, discs, ve));
			templateComp.addModel(new CatalogRowModel(person, discs, ve));
			try {templateComp.render(new DefaultViewContext(context));}
			catch (RenderException re) {if (logger.isDebugEnabled()) logger.debug("Render err:"+re);}

			//now actually render it (the DOMWriter will take care of actually
			//setting the content type)
			if (logger.isDebugEnabled()) logger.debug("Rendering document");
			new DefaultDOMWriter().write(page, context.getResponse());
	    }
	}

	/**
	 * RenderCatalogMaintHandler -
	 */
	class RenderCatalogMaintHandler extends DefaultBaseEventListener {
		public void handleViewEvent(ViewEventContext context) throws EventException, ServletException, IOException {
			//get the XMLC object
//			XMLCFactory factory = XMLCServices.getXMLCFactory(context.getConfig());
//			EditHTML page = (EditHTML) factory.create(EditHTML.class);
            EditHTML page = (EditHTML) DefaultDOMLoader.getGlobalInstance().getDOM(EditHTML.class);

/*
			//repopulate the form if possible
			EditForm ef = (EditForm) context.get(EDIT_FORM);
			if (ef!=null) {
				String id = ef.getStringVal(EditForm.ID);
	            page.getElementDiscID().setValue(id!=null && !id.equals(EditForm.INVALID_ID) ? id : "");
	            page.getElementArtist().setValue(ef.getStringVal(EditForm.ARTIST));
	            page.getElementTitle().setValue(ef.getStringVal(EditForm.TITLE));
	            page.getElementGenre().setValue(ef.getStringVal(EditForm.GENRE));
	            page.getElementLikeBox().setChecked(ef.getBooleanVal(EditForm.IS_LIKED).booleanValue());
			}

			//show any error messages
			if (logger.isDebugEnabled()) logger.debug("Checking for errors");
			ValidationException ve = (ValidationException) context.get(EDIT_ERR);
			if (ve==null) {
				if (showDebug>2) Debug.println (this, 2, "No errors found");
			    page.getElementErrorText().getParentNode().removeChild(page.getElementErrorText());
			} else {
				List errlist = ve.getExceptionList();
				if (showDebug>2) Debug.println (this, 2, errlist.size()+"error(s) found");
				StringBuffer sb = new StringBuffer(errlist.size()>1 ? "There were several errors:" : "");
				Iterator it = errlist.iterator();
				while (it.hasNext()) {
					sb.append(" "+((ValidationException) it.next()).getMessage());
				}
			    page.setTextErrorText(sb.toString());
			}
*/

			//create a template component and render it
			if (logger.isDebugEnabled()) logger.debug("Creating template component");
			Node node = page.getDocument().getElementById("EditForm");
			BTemplate templateComp = new BTemplate(new EditModel(context));
            templateComp.setView(new DefaultTemplateView(node));
			try {templateComp.render(new DefaultViewContext(context));}
			catch (RenderException re) {if (logger.isDebugEnabled()) logger.debug("Render err:"+re);}

			//now actually render it (the DOMWriter will take care of actually
			//setting the content type)
			if (logger.isDebugEnabled()) logger.debug("Rendering document");
			new DefaultDOMWriter().write(page, context.getResponse());
	    }
	}


	//------------------------------------------------------------
	//               Template Models
	//------------------------------------------------------------
	/**
	 * CatalogModel
	 */
	class CatalogModel extends AbstractTemplateModel {

		Person person = null;
		Disc[] discs = null;
		ValidationException ve = null;

		//constructor (extract stuff from context)
		public CatalogModel(Person iperson, Disc[] idiscs, ValidationException ive) {
			person = iperson;
			discs = idiscs;
			ve = ive;
		}

		//register the model by name
		public String getName() {return "Catalog";}

		//provide items by key
        public Object getItem(String key) {
            ViewContext vc = getViewContext();
			try {
				if (key.equals("Owner")) {
					return person.getFirstname()+" "+person.getLastname()+"'s Discs";
				} else if (key.equals("AddMessage")) {
					if (discs==null || discs.length<1) return vc.getTemplateNode().cloneNode(true);
					else return "";
				} else if (key.equals("Selection")) {
//					DefaultView dv = new DefaultView(vc.getTemplateNode().cloneNode(true));
					DefaultListModel dlm = new DefaultListModel();
					dlm.add(new DefaultItemMap(EditForm.INVALID_ID, "Select One"));
					if (discs!=null) for (int i=0; i<discs.length; i++) {
						Disc d = discs[i];
						DefaultItemMap dim = new DefaultItemMap(d.getHandle(), d.getArtist()+": "+d.getTitle());
						dlm.add(dim);
					}
//					BSelect bsComp = new BSelect(dlm, dv, vc);
					BSelect bsComp = new BSelect(dlm);
					bsComp.setEnabled(discs!=null && discs.length>0);
					return bsComp;
			    } else if (key.equals("AddButton")) {
                    BAction baComp = new BAction(new GetCatalogAdd());
                    baComp.setDisableBackButton(true);
				    return baComp;
			    } else if (key.equals("ChangeButton")) {
                    BAction baComp = new BAction(new GetCatalogEdit());
                    baComp.setDisableBackButton(true);
					baComp.setEnabled(discs!=null && discs.length>0);
				    return baComp;
			    } else if (key.equals("DeleteButton")) {
                    BAction baComp = new BAction(new DoCatalogDelete());
                    baComp.setDisableBackButton(true);
					baComp.setEnabled(discs!=null && discs.length>0);
				    return baComp;
			    } else if (key.equals("ExitButton")) {
                    BAction baComp = new BAction(new DoLogout());
                    baComp.setDisableBackButton(true);
				    return baComp;
/*
				} else if (key.equals("ChangeButton") ||
						   key.equals("DeleteButton")) {
					DefaultView dv = new DefaultView(vc.getTemplateNode().cloneNode(true));
					BInput biComp = new BInput(dv);
					biComp.setEnabled(discs!=null && discs.length>0);
					return biComp;
*/
				} else if (key.equals("Errors")) {
					if (ve==null) return null;
					List errlist = ve.getExceptionList();
					StringBuffer sb = new StringBuffer(errlist.size()>1 ? "There were several errors:" : "");
					Iterator it = errlist.iterator();
					while (it.hasNext()) {
						sb.append(" "+((ValidationException) it.next()).getMessage());
					}
				    return sb.toString();
				} else {
                    return super.getItem(key);
				}
			} catch (DiscRackBusinessException e) {
				return "Disc Err:"+e;
			}
		}
	}

	/**
	 * CatalogRowModel
	 */
	class CatalogRowModel extends AbstractTemplateModel implements IterativeModel {

		Person person = null;
		Disc[] discs = null;
		Disc disc = null;
		ValidationException ve = null;
		int cntr = -1;

		//constructor (extract stuff from context)
		public CatalogRowModel(Person iperson, Disc[] idiscs, ValidationException ive) {
			person = iperson;
			discs = idiscs;
			ve = ive;
		}

		//register the model by name
		public String getName() {return "CatalogRow";}

		//provide items by key
        public Object getItem(String key) {
            ViewContext vc = getViewContext();
			try {
				if (key.equals("Artist")) {
					return (disc!=null ? disc.getArtist() : "??");
				} else if (key.equals("Title")) {
					return (disc!=null ? disc.getTitle() : "??");
				} else if (key.equals("Genre")) {
					return (disc!=null ? disc.getGenre() : "??");
				} else if (key.equals("LikeIt")) {
					return ((disc!=null && disc.isLiked()) ? "Yes sirree bob!" : "No way! Blech!");
				} else {
                    return super.getItem(key);
				}
			} catch (DiscRackBusinessException e) {
				return "Disc Err:"+e;
			}
		}

		//prepare for iteration
		public void preIterate() {
			cntr = -1;
		}

		//return true if there are more rows in the data model
		public boolean hasNext() {
			return (discs!=null && cntr<(discs.length-1));
		}

		//load the next row in the model
		public void loadNext() {
			cntr++;
			disc = discs[cntr];
		}

		//cleanup after iteration
		public void postIterate() {}
	}

	/**
	 * EditModel
	 */
	class EditModel extends AbstractTemplateModel {

		ControlEventContext ec = null;
		EditForm fm = null;
		ValidationException ve = null;
		int cntr = -1;

		//constructor (extract stuff from context)
		public EditModel(ControlEventContext iec) {
			ec = iec;
			fm = (EditForm) ec.getState(EDIT_FORM);
			ve = (ValidationException) ec.getState(EDIT_ERR);
		}

		//register the model by name
		public String getName() {return "Edit";}

		//provide items by key
        public Object getItem(String key) {
            ViewContext vc = getViewContext();
			if (key.equals("ID")) {
				String val = null;
				if (fm!=null) val = fm.getStringVal(EditForm.ID);
				if (val==null) val = EditForm.INVALID_ID;
//				DefaultView dv = new DefaultView(vc.getTemplateNode().cloneNode(true));
//				return new BInput(BInput.HIDDEN, null, val, dv, vc);
				return new BInput(BInput.HIDDEN, null, val);
			} else if (key.equals("Artist")) {
				return (fm!=null ? fm.getStringVal(EditForm.ARTIST, "") : "");
			} else if (key.equals("Title")) {
				return (fm!=null ? fm.getStringVal(EditForm.TITLE, "") : "");
			} else if (key.equals("Genre")) {
				return (fm!=null ? fm.getStringVal(EditForm.GENRE, "") : "");
			} else if (key.equals("LikeIt")) {
				boolean sel = false;
				if (fm!=null) {
					try  {sel = fm.getBooleanVal(EditForm.IS_LIKED).booleanValue();}
					catch (Exception e) {}
				}
//				DefaultView dv = new DefaultView(vc.getTemplateNode().cloneNode(true));
//				BToggleButton btbComp = new BToggleButton(null, null, null, sel, dv, vc);
				BToggleButton btbComp = new BToggleButton(null, null, null, sel);
				return btbComp;
			} else if (key.equals("SaveButton")) {
                BAction baComp = new BAction(new DoCatalogUpdate());
                baComp.setDisableBackButton(true);
				return baComp;
			} else if (key.equals("CancelButton")) {
                BAction baComp = new BAction(new GetCatalog());
                baComp.setDisableBackButton(true);
				return baComp;
			} else if (key.equals("Errors")) {
				if (ve==null) return "";
				List errlist = ve.getExceptionList();
				StringBuffer sb = new StringBuffer(errlist.size()>1 ? "There were several errors:" : "");
				Iterator it = errlist.iterator();
				while (it.hasNext()) {
					sb.append(" "+((ValidationException) it.next()).getMessage());
				}
			    return sb.toString();
			} else {
                return super.getItem(key);
			}
		}
	}


	//------------------------------------------------------------
	//               HTML Form Mappings, Validators
	//------------------------------------------------------------
	/**
	 * Disc form - used to select a Disc record for change/delete
	 */
	class DiscForm extends DefaultFormMap {
		//edit form constants (these values correspond to the HTML params)
		static final String ID = "discID";

		//id values
		static final String INVALID_ID = "invalidID";

		public DiscForm() {
			//define the elements
			if (logger.isDebugEnabled()) logger.debug("Defining ID form elements");
			this.defineElement(new DefaultFormElement(ID, FormType.STRING));

			//define a form validator
			if (logger.isDebugEnabled()) logger.debug("Defining ID form validator");
			this.defineValidator(new IDValidator(ID, INVALID_ID));
		}
	}

	/**
	 * Edit form - used to actually add/change/delete a disc record
	 */
	class EditForm extends DefaultFormMap {
		//edit form constants (these values correspond to the HTML params)
		static final String ID = "discID";
		static final String ARTIST = "artist";
		static final String TITLE = "title";
		static final String GENRE = "genre";
		static final String IS_LIKED = "like";

		//id values
		static final String INVALID_ID = "invalidID";

		public EditForm() {
			//define the elements
			if (logger.isDebugEnabled()) logger.debug("Defining Edit form elements");
			this.defineElement(new DefaultFormElement(ID));
			this.defineElement(new DefaultFormElement(ARTIST,		FormType.STRING, null, new NotNullValidator("You must enter the Artist's name.")));
			this.defineElement(new DefaultFormElement(TITLE, 		FormType.STRING, null, new NotNullValidator("You must enter a Title.")));
			this.defineElement(new DefaultFormElement(GENRE, 		FormType.STRING, null, new NotNullValidator("You must enter a Genre.")));
			this.defineElement(new DefaultFormElement(IS_LIKED, 	FormType.BOOLEAN, new Boolean(false)));

			//define a form validator
			if (logger.isDebugEnabled()) logger.debug("Defining Edit form validator");
			this.defineValidator(new EditValidator());
		}
	}

	/**
	 * Edit validator - validate an edit form request
	 * based on action (add, change, or delete a record)
	 */
	class EditValidator extends DefaultFormValidator {
		public void validateForm(FormMap map, boolean deferExceptions) throws ValidationException {
			if (logger.isDebugEnabled()) logger.debug("Validating edit form");

			//get the action from the map
			Integer i = (Integer) map.getState(ACTION);
			if (i!=null) {
				//make sure we're dealing with a valid action
				int action = i.intValue();
				if (action!=ADD && action!=CHANGE && action!=DELETE) throw new ValidationException(new Integer(action), "Error Validating Edit Form: Invalid action encountered. Please try again. If the problem persists contact your system administrator.");

				//pass off request to ID validator
				new IDValidator(EditForm.ID, EditForm.INVALID_ID).validateForm(action, map, deferExceptions);
			}
		}
	}

	/**
	 * ID validator - validate an id based on action, and load
	 * the Disc it refers to if possible. This validator is a little
	 * sneaky: we reuse it for two different scenarios.
	 *
	 * When someone requests to change/delete a record, we use it to
	 * validate the ID form. In this case, the main validateForm method
	 * is invoked and we just set the action to CHANGE (regardless of
	 * what it really is...it's either that or DELETE...in either case
	 * the action is the same: make sure the record exists and load it).
	 *
	 * In the second case, we use it as an additional validation step for
	 * when someone tries to update the Edit form. In this case the second
	 * validateForm method is invoked, and the real action (ADD, CHANGE, DELETE)
	 * is passed in.
	 *
	 * In both cases, if the record exists it'll be loaded and stored in the
	 * map structure so that methods using the maps don't have to reload the
	 * Disc structure.
	 */
	class IDValidator extends DefaultFormValidator {
		String idKey = null;
		String invalidID = null;

		public IDValidator(String iidKey, String iinvalidID) {
			idKey = iidKey;
			invalidID = iinvalidID;
		}

		public void validateForm(FormMap map, boolean deferExceptions) throws ValidationException {
			validateForm (CHANGE, map, deferExceptions);
		}

		public void validateForm(int action, FormMap map, boolean deferExceptions) throws ValidationException {
			if (logger.isDebugEnabled()) logger.debug("Validating ID form");

			//make sure we've selected a valid entry
			map.putState(DISC, null);
			String id = ((DefaultFormMap) map).getStringVal(idKey);
			if ((action!=ADD) &&
                (id==null || id.equals(invalidID))) throw new ValidationException(this, "You must first select a Disc! Please try again.");

			//see if the id exists already
			boolean exists = false;
	        try {
				Disc disc = DiscFactory.findDiscByID(id);
				exists = (disc!=null);
				map.putState(DISC, disc);	//we do this so we don't have to look it up again when we get back
			} catch(Exception ex) {}

			//if action=ADD, make sure the ID doesn't already exist
			if (action==ADD && exists) throw new ValidationException(new Integer(action), "Record already exists. Please try again.");

			//if action!=ADD, make sure the ID still exists
			if (action!=ADD && !exists) throw new ValidationException(new Integer(action), "Record no longer exists (someone else may have deleted it). Please try again.");
		}
	}
}

